@@ -188,6 +188,8 @@
            #:runlines
            #:remote-test
            #:remote-exists-p
+           #:remote-exists-every-p
+           #:remote-exists-some-p
            #:remote-file-stats
            #:remote-last-reboot
            #:remote-executable-find
@@ -516,7 +518,8 @@
                              (#:file       #:consfigurator.property.file)
                              (#:os         #:consfigurator.property.os)
                              (#:service    #:consfigurator.property.service))
-           (:export #:installed
+           (:export #:known-installed-removed-packages-reset
+                    #:installed
                     #:installed-minimally
                     #:backports-installed
                     #:backports-installed-minimally
