@@ -111,18 +111,24 @@ any of the regular expressions PATTERNS."
   (:hostattrs
    (unless (or user group)
      (inapplicable-property "Not enough arguments.")))
+  ;; We don't want to execute chown(1) unless we're sure it's needed because
+  ;; in some circumstances doing so also changes the file mode.
+  (:check (multiple-value-bind (mode size mtime owner gowner)
+              (remote-file-stats path)
+            (declare (ignore mode size mtime))
+            (and (or (not user) (string= user owner))
+                 (or (not group) (string= group gowner)))))
   (:apply
-   (with-change-if-changes-file (path)
-     (if user
-         (mrun "chown" "-h" (format nil "~A~:[~;:~:*~A~]" user group) path)
-         (mrun "chgrp" "-h" group path)))))
+   (if user
+       (mrun "chown" "-h" (format nil "~A~:[~;:~:*~A~]" user group) path)
+       (mrun "chgrp" "-h" group path))))
 
 (defprop does-not-exist :posix (&rest paths)
   "Ensure that files do not exist."
   (:desc (if (cdr paths)
              #?"@{paths} do not exist"
              #?"${(car paths)} does not exist"))
-  (:check (not (apply #'remote-exists-p paths)))
+  (:check (not (apply #'remote-exists-some-p paths)))
   (:apply (mrun "rm" "-f" paths)))
 
 (defprop directory-does-not-exist :posix
@@ -243,8 +249,12 @@ directory, you can use one or more '../', or an absolute path."
        (failed-change "~A exists but is not a symbolic link." from))
      (if (and link (string= (remote-link-target from) to))
          :no-change
-         (progn
-           (containing-directory-exists from) (mrun "ln" "-sf" to from)))))
+         (progn (containing-directory-exists from)
+                ;; With at least GNU ln(1), passing -f, but not also -T, does
+                ;; not replace an existing link in some cases.
+                ;; -T is not POSIX, so we just remove any existing link first.
+                (does-not-exist from)
+                (mrun "ln" "-sf" to from)))))
   (:unapply
    (declare (ignore to))
    (if (remote-test "-L" from)
